from . import components, util
from .blueprint import simple_component

import os
import json

static_folder = simple_component.static_folder
d = lambda w: os.path.join(static_folder, w)

PRELUDE = {
    "underscore"           : d("vendor/underscore-min.js"),
    "reqwest"              : d("vendor/reqwest.min.js"),
    "EventEmitter"         : d("vendor/EventEmitter.js"),
    "pudgy/prelude"        : d("prelude.js"),
    "pudgy/loader"         : d("loader.js"),
}

PRELUDE_RAW = {
    "proxy-polyfill"       : d("vendor/proxy.min.js"),
}

PRELUDE_LINES = []

def use_jquery():
    PRELUDE["jquery"] = d("vendor/jquery-3.3.1.min.js")
    add_prelude_line("window.jQuery = require('jquery')")

def add_to_prelude(name, fname):
    PRELUDE[name] = fname

def add_prelude_line(line):
    PRELUDE_LINES.append(line)

@util.memoize
def make_prelude():
    with open(PRELUDE["pudgy/prelude"]) as f:
        loaderjs = f.read()

    out = [ loaderjs ]
    dirhash = components.CoreComponent.get_dirhash()
    out.append("require.__dirhash = '%s'" % (dirhash))

    for name in PRELUDE:
        fname = PRELUDE[name]
        if name == "pudgy/prelude":
            continue

        with open(fname) as f:

            line = f.read()
            js = json.dumps(line)
            line = """var _inj = { name: '%s', dirhash: '%s', js: %s }; define_raw(_inj.name, _inj.js, _inj.dirhash); """ % (name, dirhash, js)

            out.append(line)

    for name in PRELUDE_RAW:
        fname = PRELUDE_RAW[name]
        with open(fname) as f:
            line = f.read()
            out.append("// %s" % name)
            out.append(line)

    for line in PRELUDE_LINES:
        out.append(line)


    out.append("require('pudgy/loader')")

    return "\n".join(out)



from __future__ import print_function

from .components import *
from . import proxy

from .. import util

import flask
import dotmap

from . import assets

import os

RAPID_PUDGY_KEY="TURBO_PUDGY"
CREATE_FILES=RAPID_PUDGY_KEY in os.environ
def touch(fname):
    if os.path.exists(fname):
        os.utime(fname, None)
    else:
        open(fname, 'a').close()


def openfile(fname):
    try:
        return open(fname)
    except IOError as e:
        if CREATE_FILES:
            basedir = os.path.dirname(fname)
            try:
                os.makedirs(basedir)
            except:
                pass
            touch(fname)

            print("CREATED FILE FOR COMPONENT:", fname)
            return open(fname)
        else:
            print(" * use %s=1 to auto-create" % RAPID_PUDGY_KEY, fname)
            raise e

class Activatable(object):

    def __init__(self, *args, **kwargs):
        self.__activate_str__ = ""
        self.__activations__ = []
        super(Activatable, self).__init__(*args, **kwargs)

    def __add_activation__(self, jscode):
        self.__activations__.append(jscode)

    def __activate_tag__(self):
        self.__activate__()

        a = self.__get_activate_script__()
        if a:
            return jinja2.Markup('<script type="text/javascript">\n%s\n</script>' % a)


        return ""

    def __get_activate_script__(self):
        all = [self.__activate_str__]
        all.extend(self.__activations__)

        return "\n".join(all)

    # override this function to provide a custom activation
    def __activate__(self):
        return


class JinjaComponent(Component):
    @classmethod
    @memoize
    def get_template(cls):
        with openfile(cls.get_file_for_ext("html")) as f:
            return f.read()

    def __render__(self):
        template_str = self.get_template()
        return flask.render_template_string(template_str, **(self.context.toDict()))

class JSComponent(Activatable, Component):
    JS_LOADER=assets.JSAsset
    EXCLUDE_JS = set()
    MODULE_MAP = {}
    DEFINITIONS = {}

    @classmethod
    @memoize
    def get_js(cls):
        p = cls.get_file_for_ext(cls.JS_LOADER.EXT)
        loader = cls.get_asset_loader(p)

        dirhash_prefix = "require.__dirhash = '%s'" % cls.get_dirhash()
        with openfile(p) as f:
            l = loader.transform(f.read())

        return "%s\n%s" % (dirhash_prefix, l)

    @classmethod
    def get_asset_loader(cls, filename):
        loaders = util.inheritors(assets.AssetLoader)
        for l in loaders:
            if l.match(filename):
                return l

        return assets.JSAsset

    @classmethod
    def alias_requires(cls, name, fname):
        cls.MODULE_MAP[name] = fname

    @classmethod
    def define_requires(cls, name, filename=None, data=None):
        if filename:
            with openfile(filename) as f:
                data = f.read()


        cls.DEFINITIONS[name] = data


    @classmethod
    @memoize
    def render_requires(cls, requested, check_intersection=False):
        cls_dir = os.path.join(cls.BASE_DIR, cls.NAMESPACE)
        from .components import REQUIRE_RE

        def requires_to_js(p, basedir):
            loader = cls.get_asset_loader(p)
            sp = p
            if sp in cls.DEFINITIONS:
                js = cls.DEFINITIONS[sp]
                jsp = cls_dir
            else:
                if sp in cls.MODULE_MAP:
                    sp = cls.MODULE_MAP[p]

                jsp = loader.find_file(sp, basedir, cls_dir)
                if jsp:
                    js = loader.render_file_to_js(jsp)

            return js, jsp

        def render_requires_for_js(js, basedir):
            requires = REQUIRE_RE.findall(js)
            ret = {}
            for p in requires:
                if p in cls.EXCLUDE_JS:
                    continue

                p = p.strip("'\"")
                js, jsp = requires_to_js(p, basedir)

                if not js:
                    ret[p] = 'console.log("MISSING REQUIRE FILE %s FROM %s");' % (p, cls.__name__)
                    continue

                ret[p] = js
                ret.update(render_requires_for_js(js, os.path.dirname(jsp)))

            return ret

        def render_requires(component, basedir):
            ret = {}

            if check_intersection:
                requires = set(component.get_requires()).intersection(set(requested))
            else:
                requires = set(requested)

            for p in requires:
                js, jsp = requires_to_js(p, basedir)

                if js:
                    ret[p] = js
                    ret.update(render_requires_for_js(js, os.path.dirname(jsp)))
                else:
                    ret[p] = 'console.log("MISSING REQUIRE FILE %s FROM %s");' % (p, cls.__name__)
                    continue



            return ret

        return render_requires(cls, cls.__name__)

    def __init__(self, *args, **kwargs):
        self.__marshalled__ = False
        self.client = dotmap.DotMap()
        super(JSComponent, self).__init__(*args, **kwargs)

    def __marshal__(self):
        if not self.__marshalled__:
            flask.request.pudgy.components.add(self)
            self.__marshalled__ = True

    def __activate__(self):
        t = """$P._load("ComponentBridge", function(m) {
            m.exports.activate_component("{{__html_id__}}", "{{ __template_name__ }}", {}, {{ &__context__ }}, {{ __display_immediately__ }} )
        })"""
        rendered =  pystache.render(t, self)
        self.__activate_str__ = rendered

    def __ajax_object__(self):
        self.__activate__()
        return { "activations" : [self.__get_activate_script__()] }

    def marshal(self, **kwargs):
        self.client.update(**kwargs)
        self.__marshal__()
        return self

class CSSComponent(Component):
    CSS_LOADER=assets.CssAsset

    @classmethod
    @memoize
    def load_css(cls):
        p = cls.get_file_for_ext(cls.CSS_LOADER.EXT)
        loader = cls.get_asset_loader(p)
        css_class = "scoped_%s" % (cls.__name__)
        with openfile(p) as f:
            return loader.transform(f.read(), css_class)

    @classmethod
    @memoize
    def get_css(cls):
        return cls.add_display_rules(cls.load_css())


    @classmethod
    def get_asset_loader(cls, filename):
        loaders = util.inheritors(assets.AssetLoader)
        for l in loaders:
            if l.match(filename):
                return l

        return assets.CssAsset

    @classmethod
    def add_display_rules(cls, data):
        return "%s\n .wf_%s, .scoped_%s { display: inherit !important; } " % (data, cls.__name__, cls.__name__)


    def __init__(self, *args, **kwargs):
        super(CSSComponent, self).__init__(self, *args, **kwargs)

        if flask.request:
            if not self.__template_name__ in flask.request.pudgy.css:
                flask.request.pudgy.css.add(self.__template_name__)

class SassComponent(CSSComponent):
    CSS_LOADER=assets.SassAsset

class MustacheComponent(Component):
    @classmethod
    @memoize
    def get_template(cls):
        with openfile(cls.get_file_for_ext("mustache")) as f:
            return f.read()

    def __render__(self):
        template_str = self.get_template()
        rendered =  pystache.render(template_str, self.context)
        return rendered

mark_virtual(
    MustacheComponent,
    JSComponent,
    SassComponent,
    CSSComponent,
    JinjaComponent,
)

# A Big Package will automatically include its requires into a
# package definition
class BigJSPackage(JSComponent):
    @classmethod
    @memoize
    def get_defines(cls):
        reqs = cls.get_requires()
        return cls.render_requires(reqs)

class BigCSSPackage(Component):
    pass

mark_virtual(
    BigJSPackage,
    BigCSSPackage,
)



from .components import Component

import flask
import os

from .bigpipe import Pipeline

from .basic import RAPID_PUDGY_KEY, touch
from .basic import Activatable

from ..util import dated_url_for

import pystache

# for a Page to be a proper Component, it needs to give an ID to its body
class Page(Activatable, Pipeline):
    def __init__(self, *args, **kwargs):
        super(Page, self).__init__(*args, **kwargs)
        flask.request.pudgy.components.add(self)

        self.__marshalled__ = False

        self.__marshal__()

    def __activate__(self):
        super(Page, self).__activate__()

        # after the fact activations, set the body ID and className at the end
        # of the request. this can lead to flash of unstyled content, so prefer
        # to use FlaskPage and put the class on the body at the start
        t = 'document.body.id = "%s";' % (self.__html_id__())
        self.__add_activation__(t)

        t = 'document.body.className += " scoped_%s ";' % (self.__template_name__)
        self.__add_activation__(t)

class FlaskPage(Page, Pipeline):
    PAGE_TEMPLATE = """
<html>
<head> {{ &head }} </head>
<body id="{{ id }}" class="{{ class }}" {{ &hidden }}> {{ &body }} </body>
</html>
    """

    def __init__(self, *args, **kwargs):
        super(FlaskPage, self).__init__(*args, **kwargs)
        self.__head__ = []
        self.__stylecomps__ = []
        self.__stylesheets__ = []

    def get_head(self):
        self.render_stylesheets()
        return "\n".join(self.__head__)

    def render_stylesheet_package(self):

        b64_encode_requires = False

        if b64_encode_requires:
            import base64
            import json
            comps = json.dumps(self.__stylecomps__)
            files = json.dumps(self.__stylesheets__)

            comps = base64.b64encode(comps)
            files = base64.b64encode(files)

            url = dated_url_for('components.get_big_css', cb64=comps, fb64=files)
        else:
            url = dated_url_for('components.get_big_css',
                components=self.__stylecomps__,
                static=self.__stylesheets__)

        self.__head__.append("<link rel='stylesheet' href='%s' />" % url)

    def render_single_stylesheets(self):
        for s in self.__stylecomps__:
            url = dated_url_for('components.get_css', component=s)
            self.__head__.append("<link rel='stylesheet' href='%s' />" % url)

        for s in self.__stylesheets__:
            url = dated_url_for('static', filename=s)
            self.__head__.append("<link rel='stylesheet' href='%s' />" % url)


    def render_stylesheets(self):
        if not self.__stylecomps__ and not self.__stylesheets__:
            return [""]

        # self.render_single_stylesheets()
        self.render_stylesheet_package()

        self.__stylecomps__ = []
        self.__stylesheets__ = []

    def add_to_head(self, line):
        self.__head__.append(line)
        return self

    def add_stylesheet(self, name):
        self.__stylesheets__.append(name)
        return ""

    def add_component_stylesheet(self, c):
        self.__stylecomps__.append(c)
        return ""


    def render(self):
        self.__prepare__()

        flask.request.pudgy.components.add(self)
        kwargs = self.context.toDict()

        if RAPID_PUDGY_KEY in os.environ:
            template_dir = os.path.join(flask.current_app.root_path, flask.current_app.template_folder)
            template_file = os.path.join(template_dir, self.context.template)
            if not os.path.exists(template_file):
                try:
                    os.makedirs(os.path.dirname(template_file))
                except Exception as e:
                    pass

                touch(template_file)
                print(" * Created %s (TURBO_PUDGY)" % template_file)

        # techncially, we might render <head> twice, but i think its fine for
        # most browsers - they will treat second <head> as part of the body.
        # TODO: determine what sort of effects this can cause
        kwargs['add_stylesheet'] = self.add_stylesheet
        b = flask.render_template(self.context.template, **kwargs)
        if not self.__display_immediately__():
            self.add_component_stylesheet(self.__template_name__)

        # head after body, because we need everyone to add their styles first
        h = self.get_head()

        inst = {
            "head" : h,
            "body" : b,
            "id" : self.__html_id__(),
            "class" : self.__classname__(),
        }

        # notice that we don't call __wrap_div__ on r, because we are putting our own
        # wrappings in place
        r = pystache.render(self.PAGE_TEMPLATE, inst)
        return r



from __future__ import division, print_function


import pystache
import dotmap
import jinja2
import flask
import json
import sass
import os
import re
import hashlib


REQUIRE_RE = re.compile("""require\(['"](.*?)['"]\)""")
VIRTUAL_COMPONENTS = set()

from ..util import memoize, getrandhash, inheritors, gethash
from collections import defaultdict

def dump_values(w):
    if w:
        return w.__json__()

    return None

class Component(object):
    WRAP_COMPONENT = True
    BASE_DIR = None
    NAMESPACE = ""

    @classmethod
    def get_dirhash(c):
        return gethash(os.path.join(c.BASE_DIR, c.NAMESPACE))

    @classmethod
    def set_base_dir(cls, base_dir):
        cls.BASE_DIR = base_dir


    @classmethod
    @memoize
    def get_defines(cls):
        pass

    @classmethod
    @memoize
    def get_requires(cls):
        js = cls.get_js()
        requires = REQUIRE_RE.findall(js)

        cleaned =  [r for r in requires if r not in cls.EXCLUDE_JS]

        def clean_r(r):
            return "%s%s" % (cls.__name__, r[1:]) if r[0] == "." else r

        prefixed = [ clean_r(r)  for r in cleaned]
        return prefixed

    @classmethod
    def get_dir(cls):
        return os.path.join(cls.BASE_DIR, cls.NAMESPACE)

    @classmethod
    @memoize
    def get_file_for_ext(cls, ext):
        cls_dir = cls.get_dir()
        return os.path.join(cls_dir, cls.__name__, "%s.%s" % (cls.__name__, ext))

    @classmethod
    @memoize
    def get_css(cls):
        return ""

    @classmethod
    @memoize
    def get_js(cls):
        return ""

    @classmethod
    @memoize
    def get_js_supplements(cls):
        return []

    @classmethod
    @memoize
    def get_template(cls):
        return ""

    @classmethod
    def get_class_dependencies(cls):
        return [ ]

    @classmethod
    @memoize
    def get_version(cls):
        m = hashlib.md5()
        m.update(cls.get_package().data)

        return m.hexdigest()

    @classmethod
    @memoize
    def get_class(cls):
        return cls.__name__


    @classmethod
    @memoize
    def test_package(cls):
        if cls.__name__ in VIRTUAL_COMPONENTS:
            return

        try:
            pkg = cls.get_package()
        except Exception as e:
            print("ERROR IN PACKAGE", cls.__name__, e)
            raise e

        return


    @classmethod
    @memoize
    def get_package_object(cls):
        ret = {}
        t = cls.get_template()
        c = cls.get_css()
        j = cls.get_js()
        js_supplements = cls.get_js_supplements()
        if js_supplements:
            js_supplements = "\n".join(js_supplements)

        r = cls.get_requires()
        d = cls.get_defines()

        ret["template"] = t
        ret["css"] = c
        ret["js"] = "%s\n%s" % (j, js_supplements or "")
        ret["requires"] = r
        ret["defines"] = d

        # clean up items
        ret = {k: v for k, v in ret.items() if v}
        return ret

    @classmethod
    @memoize
    def get_package(cls):
        ret = cls.get_package_object()
        ret["namespace"] = cls.NAMESPACE
        ret["dirhash"] = cls.get_dirhash()
        return flask.jsonify(ret)

    def __init__(self, *args, **kwargs):
        self.context = dotmap.DotMap(kwargs)
        self.__template_name__ = str(self.__class__.__name__)
        self.__async__ = False
        self.__hash__ = getrandhash(self)[:10]



    def __prepare__(self):
        pass

    def __repr__(self):
        return "%s: %x" % (self.__template_name__, id(self))

    def __html_id__(self):
        return "cmp_%s" % self.__hash__

    def __activate__(self):
        return ""

    def __classname__(self):
        return "scoped_%s %s" % (self.__template_name__, self.__template_name__)

    def __json__(self):
        self.__marshal__()
        return { "_H" : self.__html_id__() }

    # return an object of this item that is suitable for transferring
    # to client.
    def __ajax_object__(self):
        return {}


    def __context__(self):
        return json.dumps(self.client.toDict(), default=dump_values)

    # should we wait for any CSS before revealing the component
    def __display_immediately__(self):
        jss = self.get_requires()
        if jss:
            for f in jss:
                if f.endswith(".css") or f.endswith(".sass"):
                    return 0

        css = self.get_css()
        if not css:
            return 1

        return 0

    def __marshal__(self):
        return self

    def __html__(self):
        self.__marshal__()
        return self.render()


    def __wrap_div__(self, div):
        if not self.WRAP_COMPONENT:
            return div

        if self.__display_immediately__():
            return "<div id='%s' class='immediate %s'>%s</div>" % (self.__html_id__(), self.__classname__(), div)

        return "<div id='%s' class='%s' style='display: none;'>%s</div>" % (self.__html_id__(),
            self.__classname__(), div)

    def __render__(self):
        return ""

    def render(self):
        if not self.__async__: # async components get prepared separately
            self.__prepare__()

        div = self.__render__()
        wrapped = self.__wrap_div__(div)
        return wrapped

class CoreComponent(Component):
    NAMESPACE='core'

def set_base_dir(d):
    Component.set_base_dir(os.path.join(d, "components"))
    CoreComponent.set_base_dir(d)

# we want to create a lookup from
# (hash(BASE_DIR), namespace) -> actual component dir
DIRS = {}
CLASSES = defaultdict(dict)
NAMESPACES = defaultdict(dict)
COMPONENT_NAMES = {}

@memoize
def validate_components():
    valid = 0
    broken = []
    virtual_components = set()

    for c in inheritors(Component):
        base_dir = c.get_dirhash()

        if not base_dir in DIRS and hasattr(c, "render_requires"):
            CLASSES[base_dir] = c
            DIRS[base_dir] = base_dir


        if c.__name__ in VIRTUAL_COMPONENTS:
            virtual_components.add(c.__name__)
        else:
            # TODO: allow multiple components to have the same name
            if c.__name__ in COMPONENT_NAMES:
                raise Exception("Declared %s but it is redefining %s from %s" %
                    (c, c.__name__, COMPONENT_NAMES[c.__name__]))
            else:
                COMPONENT_NAMES[c.__name__] = c

        try:
            pkg = c.test_package()
            if not c.__name__ in virtual_components:
                valid += 1
        except Exception as e:
            s = "%s Errors:" % (c.__name__)
            s_ = "-" *  len(s)
            broken.append(c.__name__)
            print(s)
            print(s_)
            print(e)

    print("Validated %s components before first request, %s broken" % (valid + len(broken), len(broken)))
    if broken:
        print("Broken:", ",".join(broken))


def mark_virtual(*cls):
    for c in cls:
        VIRTUAL_COMPONENTS.add(c.__name__)


mark_virtual(
    Component,
    CoreComponent,
)

def get_basedir(dirhash):
    return DIRS[dirhash]

def get_baseclass(dirhash):
    return CLASSES[dirhash]

def Virtual(cls):
    mark_virtual(cls)
    return cls



from __future__ import print_function

import jinja2
import flask
import json
import dotmap
import base64

# 2/3 python
try:
    from StringIO import BytesIO as IO
except ImportError:
    from io import BytesIO as IO

import gzip

from flask import Blueprint, render_template, abort

simple_component = Blueprint('components', __name__,
        template_folder='templates', url_prefix='/components',static_folder='static')

from . import prelude, util, components

components.set_base_dir(simple_component.root_path)

import os
import sys

from .util import memoize, dated_url_for
from .components import Component, CSSComponent


@simple_component.route('/prelude.js')
def get_prelude():
    return prelude.make_prelude()

def get_component_by_name(component):
    all_components = util.inheritors(Component)
    found = None
    for c in all_components:
        if c.__name__ == component:
            found = c
            break

    return found

@simple_component.route("/")
def index():
    pass

@simple_component.route('/<component>/invoke/<fn>', methods=['POST'])
def invoke(component, fn):
    found = get_component_by_name(component)

    js = flask.request.get_json()
    cid = js.get("cid", None)
    args = js.get("args", [])
    kwargs = js.get("kwargs", {})

    err = None
    ret = None

    res = {}
    r = {}
    try:
        ret, proxy = found.invoke(cid, fn, args, kwargs)

        r.update(proxy.__ajax_object__())
        r["response"] = ret
    except Exception as e:
        err = str(e)
        r["error"] = err
        print("ERROR INVOKING", component, fn, e)
        import traceback
        traceback.print_exc(file=sys.stdout)

    res[cid] = r

    dirhash = {}
    for p in flask.request.pudgy.components:
        res[p.__html_id__()] = p.__ajax_object__()

    res["__request__"] = {
        "css" : list(flask.request.pudgy.css),
    }

    return flask.jsonify(res)


@memoize
def build_css_package(files, components):
    all = []

    for file in files:
        filename = file.strip(".")

        with open(os.path.join(flask.current_app.static_folder, filename)) as f:
            all.append(CSSComponent.CSS_LOADER.transform(f.read()))

    for component in components:
        found = get_component_by_name(component)
        all.append(found.get_css())

    return "\n".join(all)

@simple_component.route('/pkg/<dirhash>')
def get_requires(dirhash):
    d = flask.request.args.get('data[m]')
    requested = json.loads(d)

    base_dir = components.get_basedir(dirhash)
    found = components.get_baseclass(dirhash)

    if found:
        return flask.jsonify(found.render_requires(requested))

    abort(404)

@simple_component.route('/pkg/components')
def get_component_package():
    d = flask.request.args.get("data[m]")
    cs = json.loads(d)

    ret = {}
    for c in cs:
        found = get_component_by_name(c)
        if found:
            ret[c] = found.get_package_object()


    return flask.jsonify(ret)

@simple_component.route('/csspkg')
def get_big_css():
    components = flask.request.args.getlist('components')
    files = flask.request.args.getlist('static')

    if flask.request.args.get('cb64'):
        cb64 = flask.request.args.get('cb64')
        cbs = json.loads(base64.b64decode(cb64))
        components.extend(cbs)


    if flask.request.args.get('fb64'):
        fb64 = flask.request.args.get('fb64')
        fbs = json.loads(base64.b64decode(fb64))
        files.extend(fbs)


    all = build_css_package(files, components)

    r = flask.Response(all)
    r.headers["Content-Type"] = "text/css"
    return r

    abort(404)

@simple_component.route('/<component>')
def show(component):
    found = get_component_by_name(component)

    if found:
        return found.get_package()

    abort(404)

def add_components():
    flask.request.pudgy = dotmap.DotMap()
    flask.request.pudgy.components = set()
    flask.request.pudgy.css = set()
    flask.request.pudgy.pagelets = set()

def marshal_components(prelude=True):
    from . import components
    # when lc.__html__ is called, __marshal__ is invoked, so we use lc.render()
    # instead
    lc = components.bridge.ComponentBridge()
    lc.context.components = flask.request.pudgy.components
    flask.request.pudgy.components = set()
    html = lc.render()

    # TODO: dont send same component version down multiple times
    component_versions = {}
    component_dirhashes = {}
    for c in lc.context.components:
        if c.__marshalled__:
            component_versions[c.get_class()] = "%s" % c.get_version()
            component_dirhashes[c.get_class()] = util.gethash(c.BASE_DIR)
            if c.get_class() in flask.request.pudgy.css:
                flask.request.pudgy.css.remove(c.get_class())

            for d in c.get_class_dependencies():
                component_versions[d.get_class()] = "%s" % d.get_version()
                component_dirhashes[d.get_class()] = util.gethash(d.BASE_DIR)

    component_versions[lc.get_class()] = lc.get_version()

    postfix = ""
    if flask.request.pudgy.components:
        postfix = marshal_components(prelude=False)

    css = flask.request.pudgy.css
    flask.request.pudgy.css = set()

    big_package = []
    big_package.sort(key=lambda w: w.__name__)
    for c in css:
        if type(c) == str:
            big_package.append(c)



    cb64 = base64.b64encode(json.dumps(big_package))
    return jinja2.Markup(render_template("inject_components.html",
        css_package=big_package, postfix=postfix, prelude=prelude, html=html, url_for=dated_url_for,
        versions=json.dumps(component_versions)))

def render_component(name, **kwargs):
    found = get_component_by_name(name)

    return found(**kwargs)

def inject_components(response):
    if not flask.request.pudgy.pipelined:
        if response.headers["Content-Type"].find("text/html") == 0:
            if flask.request.pudgy.components:
                injection = marshal_components()
                response.set_data("%s\n%s" % (response.get_data(as_text=True), injection))

    return response

@simple_component.before_app_first_request
def install_pudgy():
    components.validate_components()


    install(flask.current_app)

    # because our first before_request handler never runs, we invoke it
    # manually
    add_components()

# we compress our requests if we can
@simple_component.after_request
def compress_request(response):
    accept_encoding = flask.request.headers.get('Accept-Encoding', '')
    if 'gzip' not in accept_encoding.lower():
        return response

    if (response.status_code < 200 or
        response.status_code >= 300 or
        'Content-Encoding' in response.headers):
        return response

    response.direct_passthrough = False
    if len(response.data) < 2048:
        return response

    gzip_buffer = IO()
    gzip_file = gzip.GzipFile(mode='wb',fileobj=gzip_buffer)
    gzip_file.write(response.data)
    gzip_file.close()

    response.data = gzip_buffer.getvalue()
    response.headers['Content-Encoding'] = 'gzip'
    response.headers['Vary'] = 'Accept-Encoding'
    response.headers['Content-Length'] = len(response.data)

    return response

def install(app):
    app.jinja_env.trim_blocks = True
    app.jinja_env.lstrip_blocks = True

    app.before_request(add_components)
    app.after_request(inject_components)
    app.jinja_env.globals.update(CC=render_component)

def register_blueprint(app, component_dir=None):
    app.register_blueprint(simple_component)

    if not component_dir:
        component_dir = "components"

    component_dir = os.path.join(app.root_path, component_dir)


    base_dir = os.path.abspath(component_dir)
    print(" * Pudgy component dir is", base_dir)
    Component.set_base_dir(base_dir)
